/*
Utility.cpp					Copyright 2005 George A. Nelson
							Permission for use granted provided
							this notice remains intact.

By:  Tony Nelson  on   1 Mar 05

Description:  Useful utility routines.  Uses Richter's CmnHdr.h.

*/

#include "CmnHdr.h"
#include "Utility.h"
#include <WindowsX.h>
#include <tchar.h>


/*	Disable all items in (one level of) a menu.
*/
	void
DisableMenuItems( HMENU hmenu )
{
	int n = GetMenuItemCount( hmenu );
	for( int i = 0;  i < n;  ++i )
		EnableMenuItem( hmenu, i, MF_BYPOSITION | MF_GRAYED );
}


/*	Get a child window's bounding rect in the client coordinates of another
	window, such as its parent or higher ancestor.  Useful when positioning
	child windows.
	
	Oww!  Why is this so hard?  Not the client rect, not the window rect (which
	is in global coords), but the outside of the child window in the other
	window's client coords.

	GetChildRect( hwnd, hwnd, &r );		r.top == - distance from client top to
										window top.
*/
	VOID
GetChildRect( HWND hwndChild, HWND hwndParent, LPRECT prc )
{
	GetWindowRect( hwndChild, prc );
	MapWindowPoints( NULL, hwndParent, (POINT *) prc, 2 );	// ScreenToClient() could also be used
}


/*	Make a random integer 0..size.  The C Standard Libary's rand() function
	value is interpreted as the fractional part of a fixed point number.
*/
	int
RandInt( int size )
{
	unsigned short rnd = (unsigned short) rand();
	unsigned long n = (unsigned long)size * rnd;
	n >>= 8*sizeof(short) - 1;
	return n;
}


/*	Report errors via MessageBox().

	hwnd:	The window that had the error, to be disabled during MessageBox(). 
			If NULL or not a window, all app windows are disabled.
	err:	A system error code (GetLastError())
	msgIDS: Stringtable ID of a format string containing %1!d! and %2
			specifiers in any order, e.g.: "Trouble saving file, error=%1!d! %2"
	ttlIDS:	Stringtable ID of the window title for MessageBox().
*/
	void
ErrorMsgBox( HWND hwnd, LONG err, UINT msgIDS, UINT ttlIDS )
{
	static volatile LONG busy = -1;
	if( InterlockedIncrement(&busy) > 0 ) {		// multithreading reentrancy!
		InterlockedDecrement( &busy );
		return;									// drop on floor, sigh.
	}

	if( !IsWindow(hwnd) )
		hwnd = NULL;
	
	TCHAR errStr[100] = _T("");					// get system error message text
	FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL, err, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		errStr, chDIMOF(errStr), NULL );
	
	TCHAR *str = _T("(No context message), error=%1!d! %2");
	TCHAR strl[100];							// get context text
	if( hInst && LoadString(hInst, msgIDS, strl, chDIMOF(strl)) )
		str = strl;

	TCHAR msg[256];								// format the message
	LONG_PTR args[2] = { err, (LONG_PTR) errStr };
	FormatMessage( FORMAT_MESSAGE_FROM_STRING | FORMAT_MESSAGE_ARGUMENT_ARRAY,
		str, 0, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		msg, chDIMOF(msg), (va_list *) args );
	
	TCHAR *ttl = _T("Error");					// get window title
	TCHAR ttll[100];
	if( hInst && LoadString(hInst, ttlIDS, ttll, chDIMOF(ttll)) )
		ttl = ttll;

	#if 0
		// -- I find no support for doing this.
		MSG mesg;
		while( PeekMessage(&mesg, NULL, 0, 0, PM_REMOVE) ) {
			if( mesg.message == WM_QUIT ) {
				PostQuitMessage( mesg.wParam );
				InterlockedDecrement( &busy );
				return;
			}
			else if( mesg.message == WM_PAINT )		// low priority unqueued message that 
				break;								// will probably keep happening.
		}
	#endif

	MessageBox( hwnd, msg, ttl, MB_OK | MB_ICONSTOP | MB_TASKMODAL );

	InterlockedDecrement( &busy );
}

	void
CheckErrorMsgBox( HWND hwnd, LONG err, UINT msgIDS, UINT ttlIDS )
{
	if( err )
		ErrorMsgBox( hwnd, err, msgIDS, ttlIDS );
}


/*	Slow Find In String, returns position found or -1.
*/
	int
FindInStr( TCHAR *str, TCHAR *find )
{
	int last = max( 0, lstrlen(str) - lstrlen(find) );
	for( int i = 0;  i <= last;  ++i ) {
		if( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE,
			&str[i], -1, find, -1) )
		{
			return i;
		}
	}

	return -1;
}


/*	Center a window on its parent, or another window, or some rect in screen
	coords.
*/
	void
CenterWindow( HWND hwnd )
{
	CenterWindow( hwnd, GetParent(hwnd) );
}

	void
CenterWindow( HWND hwnd, HWND onW )
{
	RECT r;
	GetWindowRect( onW, &r );
	CenterWindow( hwnd, r );
}

	void
CenterWindow( HWND hwnd, RECT &r )
{
	POINT parentCtr = { (r.left+r.right)/2, (r.top+r.bottom)/2 };
	RECT ourR;
	GetWindowRect( hwnd, &ourR );
	POINT ourCtr = { (ourR.right-ourR.left)/2, (ourR.bottom-ourR.top)/2 };
	POINT p = { parentCtr.x - ourCtr.x, parentCtr.y - ourCtr.y };
	SetWindowPos( hwnd, NULL, p.x, p.y, 0, 0, SWP_NOZORDER | SWP_NOSIZE );
}
