/*
TrayButton.cpp						Copyright  2005 by George A. Nelson

By:  Tony Nelson  on  25 Jan 05

Description:  Minimize an app to either a Taskbar Button or a Taskbar
	Notification ("SysTray") icon.  Can use a Tool window to hide the Taskbar
	Button or not change the window style.  In either case, the SysTray icon
	behaves much like a Taskbar button: single left-clicks minimize or restore
	the app's window, and right-click brings up the Window ("System") menu. 
	Usage is simple; we subclass the app's window to add our behaviors. 
	Inspired by Neal Stublen's Hyperlinks.cpp on CodeGuru.com.

	Minimizing to the SysTray is appropriate for apps whose purpose is to
	change the Windows User Experience in some way, say, adding a toolbar to
	some apps, or repurposing the Command Keys at the top of some keyboards. 
	As this is an odd place to minimize to, the user should be the one to
	choose where to display the button, via a menu item or a checkbox.  That
	setting should be saved as a preference (usually in the Registry), and
	should initially be set to a normal Taskbar button.

Usage:  Call InitTrayButton() in WinMain() with the main window still not
	visible (before calling ShowWindow(iCmdShow)).  Call SetTrayButton() when
	the user changes where to minimize the app.

Discussion:  There are two common ways for an app to not have a Taskbar button:
	use a Tool window, or have the app's main window owned by a hidden window.

	Tool windows work fairly well, and give an indication that something is up
	by their appearance.  However, Tool windows can't have a minimize button,
	which makes minimizing them a bit obscure.  Changing where to minimize to
	requires changing the window style, which is a bit tricky, and the window's
	content area will change size and needs to be adjusted.  The System menu
	state is not maintained for Tool windows, possibly because it seemed
	unnecessary as they don't show an icon at the upper left corner (though a
	right-click on the title bar brings up the System menu).

	Using a hidden owner works fine, and works with many window styles, so the
	visible main window can have a minimize box.  However, changing where to
	minimize requires either making the hidden window visble, and dealing with
	it having the Taskbar button and with keeping it off-screen (and with
	multiple monitors), or required changing the window owner, something that
	works but that the Platform SDK says is impossible:

		"After creating an owned window, an application cannot transfer
		ownership of the window to another window."

		 -- Platform SDK : User Interface Services : Windows User Interface
			: Windowing : Windows : Windows Overviews : Features
			...winui/windowsuserinterface/windowing/windows/windowfeatures.htm
	
	Visual Basic users sometimes need to change the owner of a window, and I
	found some evidence that changing the owner has worked from way back.  I'm
	not certain that it will work in the future.

	Changing the owner is done in much the same way as changing the type between
	Popup and Tool:  Hide the window, call SetWindowLong(GWL_HWNDPARENT), show
	the window, plus maybe one more call show the window.  That last one keeps
	the window from disappearing the first time it is changed from Systray to
	Taskbar Button (if iCmdShow had it minimized), which I think is due to
	documented kludges in the Platform SDK for the first call to ShowWindow(). 

	TrayButton un-subclasses the window in WM_DESTROY.  If your app crashes
	when closing the window, check that other subclassers are undoing their
	thing properly (in their WM_DESTROY handlers).  Use the ASSERT in the
	WM_DESTROY handler to verify this.

	TrayButton takes over enabling the System menu, using WM_INITMENUPOPUP,
	even for the cases where the system would have done it.  There should be no
	difference between how it does it and the built-in way.

	TrayButton doesn't use global variables and should be safe to use on more
	than one window, though probably one should not do that.  It won't work
	if installed more than once on the same window, but, really, one SysTray
	icon should be enough.

	InitTrayButton() sets iCmdShow to SW_HIDE if the app is initially
	1) minimized 2) to the SysTray.

	TrayButton works with or without Unicode, however your app needs it set.  I
	assume that you have made a global preprocessor definition if you are
	using Unicode. 

	The app small icon used to be 16x16, 256 or 16 color, I think.  In WinXP it
	is normally some other size, as determined by various preference settings
	such as the height of a window title bar.  WinXP also supports "true color"
	(24 bit) icons, so it can do a better job of resizing.  It is the caller's
	responsibility to supply the desired size icon, using either LoadIcon() or
	LoadImage().
	
	I keep getting confused between minimizing and hiding and ToolWindow and
	hidden owner.  I have it all right in the code, I just get confused reading
	the code.

*/

#include <windows.h>
#include <tchar.h>
#include <windowsX.h>

#include "TrayButton.h"


#define myGUID _T("{4484953A-C640-40b2-83F7-FBA3E7AB662D}")
#define PROP_NAME _T("TrayButton") myGUID
#define MSG_NAME PROP_NAME
#define WNDCLASS_NAME PROP_NAME

#define DIMOF(Array) (sizeof(Array) / sizeof(Array[0]))


typedef struct {
	WNDPROC		oldWndProc;
	UINT		taskbarRestartMsg;
	UINT		trayMsg;
	int			iCmdShow;
	HWND		hideWnd;
	HICON		trayIcon;
	BOOL		minimizeToTray;
} TBVars;


/*	Taskbar systray ("Taskbar Message Area") notification icon, per
	Platform SDK.
*/
	static void
AddTaskbarIcon( HWND hwnd, TBVars * tbv )
{
	NOTIFYICONDATA nid = { 0 };
	nid.cbSize           = NOTIFYICONDATA_V1_SIZE;
	nid.hWnd             = hwnd;
	nid.uFlags           = NIF_ICON | NIF_TIP | NIF_MESSAGE;
	nid.uCallbackMessage = tbv->trayMsg;
	nid.hIcon            = (HICON) SendMessage( hwnd, WM_GETICON, ICON_SMALL, 0 );
	if( !nid.hIcon )
		nid.hIcon        = (HICON) SendMessage( hwnd, WM_GETICON, ICON_BIG, 0 );
	GetWindowText( hwnd, nid.szTip, DIMOF(nid.szTip) );
	Shell_NotifyIcon( NIM_ADD, &nid );
}


	static void
RemoveTaskbarIcon( HWND hwnd )
{
	NOTIFYICONDATA nid = { 0 };
	nid.cbSize = NOTIFYICONDATA_V1_SIZE;
	nid.hWnd = hwnd;
	Shell_NotifyIcon( NIM_DELETE, &nid );
}


	static void
ShowTaskbarIconMenu( HWND hwnd )
{
	HMENU popM = GetSystemMenu( hwnd, FALSE );
	//EnableSystemMenuItems( hwnd, popM );
	SendMessage( hwnd, WM_INITMENUPOPUP, (WPARAM) popM, MAKELPARAM(0, TRUE) );

	SetForegroundWindow( hwnd );

	//DWORD at = GetMessagePos();	-- doesn't work for systray icons
	POINT at;
	GetCursorPos( &at );
	BOOL id = TrackPopupMenu( popM, TPM_RIGHTBUTTON | TPM_RETURNCMD, at.x, at.y, 0, hwnd, NULL);
	if( id )
		SendMessage( hwnd, WM_SYSCOMMAND, id, 0xFFFF0000 );
	else {
		/*	Send recommended NIM_SETFOCUS.  Although it is defined only in
			_WIN32_IE, it seems to work with NOTIFYICONDATA_V1_SIZE, and I
			expect that it will do nothing harmful on earlier versions of the
			Shell.
		*/ 
		NOTIFYICONDATA nid = { 0 };
		nid.cbSize = NOTIFYICONDATA_V1_SIZE;
		nid.hWnd = hwnd;
		Shell_NotifyIcon( 0x00000003, &nid );	//NIM_SETFOCUS
	}

	PostMessage( hwnd, WM_NULL, 0, 0 );
}


/*	The System Menu isn't automatically properly enabled when the window doesn't
	have a window icon where the System Menu pops up from, as in a Tool window,
	and it certainly isn't enabled when there is a click on the Tray Button, so
	we have to do it ourselves.  I hope this is done right.  It should exactly
	mimic the normal enabling.  I wish there were a simple call or message to
	do this; WM_INITMENUPOPUP doesn't have useful default behavior.
*/
	static void
EnableSystemMenu( HWND hwnd, HMENU hmenu, TBVars * tbv )
{
	BOOL min    = (tbv->minimizeToTray) ? !IsWindowVisible(hwnd) : IsIconic(hwnd);
	BOOL max    = !min && IsZoomed( hwnd );
	BOOL res    = !(min || max);
	DWORD style = GetWindowStyle( hwnd );

	UINT resable  = MF_ENABLED;
	UINT moveable = MF_GRAYED;
	UINT sizeable = MF_GRAYED;
	UINT minable  = (style & WS_MINIMIZEBOX) ? MF_ENABLED : MF_GRAYED;
	UINT maxable  = (style & WS_MAXIMIZEBOX) ? MF_ENABLED : MF_GRAYED;

	if( res ) {
		resable      = MF_GRAYED;
		moveable     = MF_ENABLED;
		if( style & WS_SIZEBOX )
			sizeable = MF_ENABLED;
	}
	else if( min )
		minable = MF_GRAYED;
	else if( max )
		maxable = MF_GRAYED;
	
	EnableMenuItem( hmenu, SC_RESTORE,  MF_BYCOMMAND | resable );
	EnableMenuItem( hmenu, SC_MOVE,     MF_BYCOMMAND | moveable  );
	EnableMenuItem( hmenu, SC_SIZE,     MF_BYCOMMAND | sizeable );
	EnableMenuItem( hmenu, SC_MINIMIZE, MF_BYCOMMAND | minable );			
	EnableMenuItem( hmenu, SC_MAXIMIZE, MF_BYCOMMAND | maxable );
	EnableMenuItem( hmenu, SC_CLOSE,    MF_BYCOMMAND | MF_ENABLED  );
	SetMenuDefaultItem( hmenu, SC_CLOSE, false );
	//pass to next WNDPROC
}


/*	EnumWindows callback to get the topmost application, if any.
*/
typedef struct {
	HWND	ourW;
	HWND	foundW;
	RECT	screenR;
} EnumTopmostAppData;

		static void
	GetScreenRect( RECT * rp )
	{
		rp->right  = GetSystemMetrics( SM_CXVIRTUALSCREEN );
		rp->bottom = GetSystemMetrics( SM_CYVIRTUALSCREEN );
		if( !rp->right || !rp->bottom ) {
			rp->right  = GetSystemMetrics( SM_CXSCREEN );
			rp->bottom = GetSystemMetrics( SM_CYSCREEN );
		}
		else {
			rp->left   = GetSystemMetrics( SM_XVIRTUALSCREEN );
			rp->top    = GetSystemMetrics( SM_YVIRTUALSCREEN );
			rp->right  += rp->left;
			rp->bottom += rp->top;
		}

	}

	static BOOL CALLBACK
EnumTopmostApp( HWND hwnd, LONG dl )
{
	EnumTopmostAppData *dp = (EnumTopmostAppData *) dl;

	if( !dp->screenR.right )
		GetScreenRect( &dp->screenR );

	RECT r, r1;
	GetWindowRect( hwnd, &r );
	if( !IsWindowVisible(hwnd)
	 || IsIconic(hwnd)
	 || !IntersectRect(&r1, &r, &dp->screenR) )
	{
		return TRUE;
	}

	if( hwnd == dp->ourW
	 || !(GetWindowExStyle(hwnd) & WS_EX_TOPMOST) )
	{
		dp->foundW = hwnd;
		return FALSE;
	}
	
	return TRUE;
}


/*	Window subclass:
*/
	static LRESULT CALLBACK
TrayBtnWndProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	TBVars *tbv = (TBVars *) GetProp( hwnd, PROP_NAME );
	WNDPROC oldWndProc = tbv->oldWndProc;

	switch( uMsg ) {
	case WM_SYSCOMMAND:
	{
		UINT cmd = (UINT)(wParam);
		//int x = (int)(short)LOWORD(lParam);
		//int y = (int)(short)HIWORD(lParam);

		if( tbv->minimizeToTray )
		{
			switch( cmd & 0xFFF0 ) {
			case SC_MINIMIZE:
				ShowWindow( hwnd, SW_HIDE );
				return 0;

			case SC_RESTORE:
				if( !IsWindowVisible(hwnd) )
					ShowWindow( hwnd, SW_SHOW );
				else if( IsZoomed(hwnd) )
					ShowWindow( hwnd, SW_SHOWNORMAL );
				return 0;
			}
		}
		break;
	}

	case WM_INITMENUPOPUP:
	{
		HMENU hmenu = (HMENU)(wParam);
		//UINT item = (UINT)LOWORD(lParam);
		BOOL fSystemMenu = (BOOL)HIWORD(lParam);

		if( fSystemMenu ) {
			EnableSystemMenu( hwnd, hmenu, tbv );
			//also pass to next WNDPROC
		}
		break;
	}

	case WM_DESTROY:
	{
		//ASSERT( GetWindowLong(hwnd, GWL_WNDPROC) == (LONG) TrayBtnWndProc );
		RemoveTaskbarIcon( hwnd );
		RemoveProp( hwnd, PROP_NAME );
		SetWindowLong( hwnd, GWL_WNDPROC, (LONG) tbv->oldWndProc );
		if( tbv->hideWnd ) {
			SetWindowLong( hwnd, GWL_HWNDPARENT, 0 );
			DestroyWindow( tbv->hideWnd );
		}
		if( tbv->trayIcon )
			DestroyIcon( tbv->trayIcon );
		HeapFree( GetProcessHeap(), 0, tbv );
		break;
	}

	default:
		if( uMsg == tbv->taskbarRestartMsg ) {
			AddTaskbarIcon( hwnd, tbv );
			break;
		}
		if( uMsg == tbv->trayMsg ) {
			switch( lParam ) {
			case WM_LBUTTONDOWN:
			case WM_LBUTTONDBLCLK:
				if( !IsWindowVisible(hwnd) ) {
					ShowWindow( hwnd, SW_SHOW );
					SetForegroundWindow( hwnd );
				}
				else {
					EnumTopmostAppData d = { hwnd };
					EnumWindows( EnumTopmostApp, (LONG) &d );

					if( d.foundW != hwnd )
						SetForegroundWindow( hwnd );
					else if( IsWindowEnabled(hwnd) )		// not when a dialog is up
						ShowWindow( hwnd, SW_HIDE );
				}
				break;

			case WM_RBUTTONDOWN:
				ShowTaskbarIconMenu( hwnd );
				break;
			}
			return 0;	// eat this message
		}
		break;
	}

	return CallWindowProc( oldWndProc, hwnd, uMsg, wParam, lParam );
}


/*	External interface:
*/
	BOOL /* success */
InitTrayButton( HWND hwnd, int * iCmdShowp, BOOL minimizeToTray, BOOL useToolKind )
{
	WNDPROC oldWndProc = (WNDPROC) GetWindowLong( hwnd, GWL_WNDPROC );
	if( oldWndProc == TrayBtnWndProc
	 || GetProp(hwnd, PROP_NAME) )
	{
		SetLastError( ERROR_ALREADY_INITIALIZED );
		return FALSE;
	}

	TBVars *tbv = (TBVars *) HeapAlloc( GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(TBVars) );
	if( !tbv ) {
		SetLastError( ERROR_NOT_ENOUGH_MEMORY );
		return FALSE;
	}
	
	HINSTANCE hInst = GetWindowInstance( hwnd );

	tbv->trayMsg    = RegisterWindowMessage( MSG_NAME );
	tbv->taskbarRestartMsg = RegisterWindowMessage( _T("TaskbarCreated") );
	tbv->oldWndProc = SubclassWindow( hwnd, TrayBtnWndProc );

	if( !useToolKind ) {
		WNDCLASS wc = { 0 };
		wc.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1);
		wc.hInstance     = hInst;
		wc.lpfnWndProc   = DefWindowProc;
		wc.lpszClassName = WNDCLASS_NAME;
		RegisterClass( &wc );
		tbv->hideWnd = CreateWindow( wc.lpszClassName, _T(""), WS_POPUP,
			0, 0, 0, 0, NULL, NULL, wc.hInstance, NULL ); 
	}

	SetProp( hwnd, PROP_NAME, (HANDLE) tbv );

	SetTrayButton( hwnd, minimizeToTray );

	if( minimizeToTray ) {
		switch( *iCmdShowp ) {
		case SW_MINIMIZE:
		case SW_SHOWMINNOACTIVE:
		case SW_SHOWMINIMIZED:
		case SW_FORCEMINIMIZE:
		//? case SW_SHOWMAXIMIZED:
			tbv->iCmdShow = *iCmdShowp;
			*iCmdShowp = SW_HIDE;
			break;
		}
	}
	
	return TRUE;
}


	BOOL /* success */
SetTrayButton( HWND hwnd, BOOL minimizeToTray )
{
	TBVars *tbv = (TBVars *) GetProp( hwnd, PROP_NAME );
	if( !tbv ) {
		SetLastError( ERROR_INVALID_FUNCTION );
		return FALSE;
	}

	tbv->minimizeToTray = minimizeToTray;

	BOOL vis = IsWindowVisible( hwnd );
	
	if( tbv->hideWnd ) { 
		if( vis )
			ShowWindow( hwnd, SW_HIDE );
		SetWindowLong( hwnd, GWL_HWNDPARENT, (LONG) (minimizeToTray ? tbv->hideWnd : NULL) );
		if( vis ) {
			ShowWindow( hwnd, SW_SHOW );

			/*	Without this, if the app started with the window initially
				minimized to the Systray, so that we didn't call ShowWindow with
				iCmdShow's value, and we're now removing the Owner so there will
				now be a Taskbar button, the window will minimize (or hide?) at the
				call to ShowWindow(SW_SHOW) above.  I think this is a side effect
				of kludges for app startup described in the Platform SDK docs for
				ShowWindow(). 
				
				Note that using SW_SHOWNORMAL in the ShowWindow() call above does
				not keep the window from hiding; this second call is needed.  I
				think the Taskbar button is made when we remove the owner, and
				that this call must be made after the taskbar button is created for
				the first time.  Subsequent calls that make the Taskbar button
				don't have this problem.  Using a Tool window doesn't seem to need
				this.  Yow.
			*/
			if( tbv->iCmdShow ) {
				ShowWindow( hwnd, SW_SHOWNORMAL );
				tbv->iCmdShow = 0;
			}
		}
	}
	else {	// !tbv->hideWnd
		LONG style = GetWindowExStyle( hwnd );
		style &= ~WS_EX_TOOLWINDOW;
		if( minimizeToTray )
			style |= WS_EX_TOOLWINDOW;
		
		RECT r;
		GetClientRect( hwnd, &r );
		
		if( vis )
			ShowWindow( hwnd, SW_HIDE );
		SetWindowLong( hwnd, GWL_EXSTYLE, style );
		if( vis )
			ShowWindow( hwnd, SW_SHOW );
		
		AdjustWindowRectEx( &r, GetWindowStyle(hwnd), FALSE, GetWindowExStyle(hwnd) );
		SetWindowPos( hwnd, NULL, 0, 0, r.right - r.left, r.bottom - r.top,
			SWP_NOZORDER | SWP_NOMOVE );
	}

	if( minimizeToTray )
		AddTaskbarIcon( hwnd, tbv );
	else
		RemoveTaskbarIcon( hwnd );

	return TRUE;
}
